(***********************************************************************)
(*                                                                     *)
(*                          CamlP3l                                    *)
(*                                                                     *)
(* (C) 2004-2007                                                       *)
(*             Roberto Di Cosmo (dicosmo@dicosmo.org)                  *)
(*             Zheng Li (li@pps.jussieu.fr)                            *)
(*             Pierre Weis (Pierre.Weis@inria.fr)                      *)
(*             Francois Clement (Francois.Clement@inria.fr)            *)
(*                                                                     *)
(* Based on original Ocaml P3L System                                  *)
(* (C) 1997 by                                                         *)
(*             Roberto Di Cosmo (dicosmo@ens.fr)                       *)
(*             Marco Danelutto (marcod@di.unipi.it)                    *)
(*             Xavier Leroy  (Xavier.Leroy@inria.fr)                   *)
(*             Susanna Pelagatti (susanna@di.unipi.it)                 *)
(*                                                                     *)
(* This program is free software; you can redistribute it and/or       *)
(* modify it under the terms of the GNU Library General Public License *)
(* as published by the Free Software Foundation; either version 2      *)
(* of the License, or (at your option) any later version.              *)
(*                                                                     *)
(* This program is distributed in the hope that it will be useful,     *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of      *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *)
(* GNU Library General Public License for more details.                *)
(*                                                                     *)
(***********************************************************************)

(* $Id: toolbase.ml,v 1.14 2007/07/13 13:49:22 fclement Exp $ *)

(* Reserved keys *)
let keys = [
  "seq"; "-sequential";
  "gra"; "-graphical";
  "par"; "-parallel";
  "pro"; "-process";
  "thr"; "-thread";
  "gra"; "-graphical";
  "v"; "-version";
  "h"; "help"; "-help";
  "dp3"; "-debug-p3ml";
];;

type exe =
   | Byt | Bin | Top;;
type semantic =
   | Seq | Par_pro | Par_thr | Gra ;;

let caml_dir = Installation_data.caml_library_path;;
let p3ml_dir = Installation_data.camlp3l_library_path;;

let make_caml_subdir dir = Filename.concat caml_dir dir;;
let make_p3ml_subdir dir = Filename.concat p3ml_dir dir;;

let ulib_dir = make_caml_subdir "ulib"
and p3ml_seq_dir = make_p3ml_subdir "seq"
and p3ml_gra_dir = make_p3ml_subdir "gra"
and p3ml_par_dir = make_p3ml_subdir "par";;

(* Semantics evaluation association list =
   sem * (sem_name * sem_path_list * sem_ext_lib) *)
let sem_assoc = [
  Seq, ("sequential",
        [p3ml_seq_dir],
        ["unix"; "bigarray"; "ulib"; "p3l";]);
  Gra, ("graphical",
        [p3ml_gra_dir],
        ["graphics"; "unix"; "bigarray"; "ulib"; "p3l";]);
  Par_pro, ("parallel (with processes)",
            [Filename.concat p3ml_par_dir "pro"],
            ["unix"; "bigarray"; "ulib"; "p3l";]);
  Par_thr, ("parallel (with threads)",
            [Filename.concat p3ml_par_dir "thr"; "+threads"],
            ["threads"; "unix"; "bigarray"; "ulib"; "p3l";]);
];;

(* Executable kind association list =
   exe * (exe_name * exe_cmd * exe_sfx_obj * exe_sfx_lib) *)
let exe_assoc = [
  Byt, ("CamlP3l bytecode compiler", "ocamlc", "cmo", "cma");
  Bin, ("CamlP3l nativecode compiler", "ocamlopt", "cmx", "cmxa");
  Top, ("CamlP3l toplevel (bytecode)", "ocaml", "cmo", "cma");
];;

(* Configuration *)
type config = {
  mutable exe : exe;
  mutable sem : semantic;
  mutable args : string list;
  mutable path : string;
  mutable debug : bool;
};;

(* default configuration *)
let default_configuration = {
  exe = Byt;
  sem = Seq;
  args = [];
  path = p3ml_seq_dir;
  debug = false;
};;

(* Setting the semantics evaluation regime choice. *)
let set_seq () = default_configuration.sem <- Seq;;
let set_gra () = default_configuration.sem <- Gra;;
let set_par_pro () = default_configuration.sem <- Par_pro;;
let set_par_thr () = default_configuration.sem <- Par_thr;;
let set_debug () = default_configuration.debug <- true;;

(* Bulding the command line options for the semantics evaluation regime. *)
let make_seq_option opt_name =
  (opt_name, Arg.Unit set_seq, " Use sequential execution");;
let make_gra_option opt_name =
  (opt_name, Arg.Unit set_gra, " Use graphical execution");;
let make_par_pro_option opt_name =
  (opt_name, Arg.Unit set_par_pro,
   " Use parallel execution with process model");;
let make_par_thr_option opt_name =
  (opt_name, Arg.Unit set_par_thr,
   " Use parallel execution with thread model");;
let make_debug_option opt_name =
  (opt_name, Arg.Unit set_debug, " Debug option for the CamlP3l compiler");;

(* Instantiate configuration *)
let set_args () =
  let argv_m =
    Array.of_list
      (List.fold_left
         (fun l s ->
           if String.get s 0 = '-' then
             let key = String.sub s 1 (String.length s - 1) in
             if not (List.mem key keys) then l @ ["-ext"; key] 
             else l @ [s]
           else l @ [s])
         []
         (Array.to_list Sys.argv)) in
  let speclist =
    [
     make_seq_option "-seq"; make_seq_option "--sequential";
     make_gra_option "-gra"; make_gra_option "--graphical";
     make_par_pro_option "-par"; make_par_pro_option "--parallel";
     make_par_pro_option "-pro"; make_par_pro_option "--process";
     make_par_thr_option "-thr"; make_par_thr_option "--thread";
     make_debug_option "-dp3"; make_debug_option "--debug-p3l";
     ("-ext",
      Arg.String
        (fun s ->
          default_configuration.args <-
            default_configuration.args @ ["-" ^ s]),
      " Internal use only")
   ] in
  try
    Arg.parse_argv
      argv_m
      speclist
      (fun s -> default_configuration.args <- default_configuration.args @ [s])
      "CamlP3l can also accept most compiler options of OCaml."
  with
  | Arg.Bad msg -> Printf.eprintf "%s" msg; exit 2;
  | Arg.Help msg -> Printf.printf "%s" msg; exit 0;;

(* Add suffix string [suff] to filename [fname]. *)
let add_suffix suff fname = Printf.sprintf "%s.%s" fname suff;;

(* A command is constructed by the following parts in the following
   sequence:
   Cmd(exe) + ModOpt(model) + ModPath(model) + ModLib(model+exe) +
   SemLib(sem+exe) + SemObj(sem+exe) + UsrCmd *)
let get_cmd () =

  let exe_name, exe_cmd, exe_sfx_obj, exe_sfx_lib =
    List.assoc default_configuration.exe exe_assoc in

  let sem_name, sem_path_list, sem_ext_lib =
    List.assoc default_configuration.sem sem_assoc in

  let compile_only = List.mem "-c" default_configuration.args in

  let sem_path_list =
    (* When compiling only, we just need the root directory where camlp3l is
       installed, otherwise we need to access with the p3l.cm*a library from
       the relevant sub-directory for the given semantics. *)
    if compile_only then [p3ml_dir; ulib_dir] else
    p3ml_dir :: ulib_dir :: sem_path_list in
  let all_inc_path =
    List.flatten (List.map (fun s -> ["-I"; s]) sem_path_list) in 
  let all_ext_lib =
    if compile_only then [] else
    List.map (add_suffix exe_sfx_lib) sem_ext_lib in
  let argv_lib = all_inc_path @ all_ext_lib in
  let argv_usr =
    let argv_usr_dup_removed =
      List.filter
        (fun x -> not (List.mem x all_ext_lib))
        default_configuration.args in
    List.fold_left
      (fun l x ->
         if Filename.check_suffix x "3l"  then
           l @ ["-intf-suffix .3li -impl"; x] else
         if Filename.check_suffix x "3li" then l @ ["-intf"; x] else
         l @ [x]
      ) [] argv_usr_dup_removed in
  let argv_dst =
    let rec guess_exe_name = function
      | [] -> None
      | h :: _ when
          List.exists (Filename.check_suffix h) ["cmo"; "cmx"] ->
          Some (Filename.chop_extension h)
      | h1 :: h2 :: _ when h2 = "-impl" ->
          Some (Filename.chop_extension h1)
      | _ :: t -> guess_exe_name t in
    let exe_sfx_dst = "opt" in
    let sem_sfx_dst = String.sub sem_name 0 3 in
    if List.exists (fun x -> x = "-c" || x = "-o") argv_usr then []
    else match default_configuration.exe with
      | Top -> []
      | Byt ->
          (match guess_exe_name (List.rev argv_usr) with
          | None -> []
          | Some x -> ["-o"; add_suffix sem_sfx_dst x])
      | Bin ->
          (match guess_exe_name (List.rev argv_usr) with
           | None -> []
           | Some x ->
               ["-o"; String.concat "." [x; sem_sfx_dst; exe_sfx_dst]])
  in
  let desc =
    Printf.sprintf "Launching %s with %s semantics." exe_name sem_name in
  (exe_cmd, argv_lib @ argv_usr @ argv_dst, desc)
;;

let make_open_cmd mods =
  List.map (fun x -> Printf.sprintf "open %s\\;\\;" x) mods;;

let add_version_options bn =
  let si =
    {Version.software_info with Uversionning.software_banner_name = bn} in
  Uversionning.add_version_options si
;;

let make_main exe bn =
  add_version_options bn;
  let command, arguments, description =
    set_args ();
    default_configuration.exe <- exe;
    get_cmd () in

(*  let open_cmd = make_open_cmd ["P3l"] in

  let preproc_cmd =
    ["-pp "; "\"echo -n " ^ String.concat "" open_cmd ^ " | cat - \""] in

  let cmd = String.concat " " (command :: preproc_cmd @ arguments) in
*)

  let cmd = String.concat " " (command :: arguments) in

  let main () =
    if default_configuration.debug then begin
       prerr_endline description;
       prerr_endline (Printf.sprintf "Launching %S" cmd);
    end;
    exit (Sys.command cmd) in
  main;;

(*****
(** Other useful tools. *)

let get_path_list path =
  let rec split s =
    try
      let pos = String.index s ':' in
      String.sub s 0 pos ::
      split (String.sub s (pos + 1) ((String.length s) - pos - 1)) with
    | Not_found -> [s] in
  try split (Sys.getenv path) with
  | Not_found -> [];;
(** Get the user's system environment variable env. *)
*******)

let make_top_main exe bn =
  add_version_options bn;
  let command, arguments, description =
    set_args ();
    default_configuration.exe <- exe;
    get_cmd () in

  let init_file_name = Filename.temp_file "camlp3l" ".ml" in

  let build_init_file () =
    let oc = open_out init_file_name in
    List.iter (fun m ->
      output_string oc (Printf.sprintf "open %s;;\n" m))
      ["P3l"];
    output_string oc (Printf.sprintf "Sys.remove %S;;\n" init_file_name);
    close_out oc in

  let cmd =
    String.concat " " ([command; "-init"; init_file_name] @ arguments) in

  let main () =
    build_init_file ();
    if default_configuration.debug then begin
      prerr_endline description;
       prerr_endline (Printf.sprintf "Launching %S" cmd);
    end;
    exit (Sys.command cmd) in

   main;;
