#!/bin/sh

#######################################################################
#                                                                     #
#                          CamlP3l                                    #
#                                                                     #
# (C) 2004-2007                                                       #
#             Roberto Di Cosmo (dicosmo@dicosmo.org)                  #
#             Zheng Li (li@pps.jussieu.fr)                            #
#             Pierre Weis (Pierre.Weis@inria.fr)                      #
#             Francois Clement (Francois.Clement@inria.fr)            #
#                                                                     #
# Based on original Ocaml P3L System                                  #
# (C) 1997 by                                                         #
#             Roberto Di Cosmo (dicosmo@ens.fr)                       #
#             Marco Danelutto (marcod@di.unipi.it)                    #
#             Xavier Leroy  (Xavier.Leroy@inria.fr)                   #
#             Susanna Pelagatti (susanna@di.unipi.it)                 #
#                                                                     #
# This program is free software; you can redistribute it and/or       #
# modify it under the terms of the GNU Library General Public License #
# as published by the Free Software Foundation; either version 2      #
# of the License, or (at your option) any later version.              #
#                                                                     #
# This program is distributed in the hope that it will be useful,     #
# but WITHOUT ANY WARRANTY; without even the implied warranty of      #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       #
# GNU Library General Public License for more details.                #
#                                                                     #
#######################################################################

# $Id: mk_config_files.sh,v 1.3 2007/02/16 16:10:20 fclement Exp $

SCRIPT=`echo $0 | awk -F/ '{print $NF}'`

EXIT=0
DEBUG=no
HELP=no
FORCE=no
DEFAULTNAME=yes
EXIST=no

baseresname=RES
output=none
Nit=-1
step=4
sleep=5

[ $# -eq 0 ] && HELP=yes && EXIT=2
while [ $# -ge 1 ]; do
    case $1 in
        -x) set -x;;
        -d) DEBUG=yes;;
        -h|-help|--help) HELP=yes;;
        -f) FORCE=yes;;
        -nx|-ny|-n) shift; N=$1;;
        -nsbdx|-nsbdy|-nsbd) shift; Nsbd=$1;;
        -name) shift; DEFAULTNAME=no; name=$1;;
        -baseresname|-basename|-bn) shift; baseresname=$1;;
        -output|-o) shift; output=$1;;
        -Nit) shift; Nit=$1;;
        -step) shift; step=$1;;
        -sleep) shift; sleep=$1;;
        -*) echo "Unknown option $1"; HELP=yes; EXIT=2;;
        *) echo "Unused argument $1, skipped...";;
    esac
    shift
done

[ -z "$N" -o -z "$Nsbd" ] && HELP=yes && EXIT=2

if [ $HELP = "yes" ]; then
    echo
    echo "Usage: $SCRIPT [<options>] -n <n> -nsbd <nsbd>"
    echo
    echo "Generate the configuration files for running the coupling example"
    echo "and display the corresponding complexity."
    echo "If the configuration files already exist, they are simply displayed,"
    echo "unless the option -f is provided."
    echo ""
    echo "ARGUMENTS"
    echo "  -n <n>: set the number of grid points in each direction to <n>."
    echo "  -nsbd <n>: set the number of subdomains in each direction to <n>."
    echo
    echo "OPTIONS"
    echo "  -f: force the (re)generation of the configuration files."
    echo "  -name <name>: generate files \"config<name>\" and \"rdec2<name>.cfg\"."
    echo "                Default is \"-<nsbd>x<nsbd>-<n>x<n>\"."
    echo "  -bn <name>: set the basename for the numerical results to <name>."
    echo "              Default is \"$baseresname\"."
    echo "  -o <type>: set the output to <type>."
    echo "             Default is \"$output\". Valid entries are:"
    echo "                 \"window\" for graphical display (needs gnuplot)"
    echo "                 \"ps\" for postscript output (in directory \"res\")"
    echo "                 \"none\" for performance measurements."
    echo "  -Nit <n>: set the number of iterations to <n>."
    echo "            Default is automatically computed for correct convergence."
    echo "  -step <n>: set the increment step for graphical outputs to <n>."
    echo "             Default is $step, or the value set by -Nit option if it is used."
    echo "  -sleep <n>: set the amount of time in seconds during which the \"window\""
    echo "              remains displayed after the computation."
    echo "              Default is $sleep."
    echo
    exit $EXIT
fi

if [ $DEFAULTNAME = "yes" ]; then
    name=-${Nsbd}x${Nsbd}-${N}x${N}
fi
config=config$name
rdec=rdec2$name.cfg

n=$(($N / $Nsbd))
p=$(($n / 5))
[ $p -eq 0 ] && p=1

if [ $Nit -le 0 ]; then
    case $Nsbd in
        1) Nit=5;;
        2) Nit=10;;
        *) Nit=$(((5 * $Nsbd - 1) * $Nsbd / 2 + 4));;
    esac
else
    config=${config}_$Nit
    step=$Nit
fi
nit=$(($n / 3 * 10))
[ $nit -eq 0 ] && nit=10

CONFIG="\
rdec_config_filename = \"$rdec\";
result_basename = \"$baseresname\";
graphical_output = \"$output\";
outer_itermax = $Nit;
inner_itermax = $nit;
display_step = $step;
a_while = $sleep;"

RDEC="\
x0 = 0.; y0 = 0.;
lx = 1.; ly = 1.;
nx = $N; ny = $N;
px = $p; py = $p;
nsbdx = $Nsbd; nsbdy = $Nsbd;"

if [ $FORCE = "no" -a -f $config ]; then
    EXIST=yes
    echo
    echo "File $config already exists:"
    cat $config
else
    echo "$CONFIG" > $config
fi
if [ $FORCE = "no" -a -f $rdec ]; then
    EXIST=yes
    echo
    echo "File $rdec already exists:"
    cat $rdec
else
    echo "$RDEC" > $rdec
fi

simplify () {
    COMP=$1
    UNIT=$2
    while [ $COMP -gt 10000 -a $UNIT != "Pflop" ]; do
        COMP=$(($COMP / 1000))
        case "$UNIT" in
            flop) UNIT=kflop;;
            kflop) UNIT=Mflop;;
            Mflop) UNIT=Gflop;;
            Gflop) UNIT=Tflop;;
            Tflop) UNIT=Pflop;;
            *) COMP=$(($COMP * 1000));;
        esac
    done
    echo $COMP@$UNIT
}

comp=$((147 * $N * $N))
unit=flop
compunit=`simplify $comp $unit`
comp=`echo $compunit | awk -F@ '{print $1}'`
unit=`echo $compunit | awk -F@ '{print $2}'`
comp=$(($comp * $N * $Nsbd))
compunit=`simplify $comp $unit`
comp=`echo $compunit | awk -F@ '{print $1}'`
unit=`echo $compunit | awk -F@ '{print $2}'`
echo
echo "The approximated complexity, i.e. the number of floating point operations,"
echo "of the corresponding coupling example will be: $comp $unit."
echo

if [ $EXIST = "yes" ]; then
    echo "To regenerate already existing files, use the force option -f."
    echo
fi

exit $EXIT
