(* $Id: make_coupling.ml,v 1.5 2007/07/13 13:49:22 fclement Exp $ *)

(*****************************************************************************)
(* IMPORTANT:                                                                *)
(* - remember to fflush data output in the C program, or you will block!     *)
(* - remember not to rely on global references/arrays; thus it is better to  *)
(*   put them inside the data stream.                                        *)
(* - Meta-data, such as the number of workers for farm or mapvector          *)
(*   skeletons, must be set BEFORE the computation network is deployed,      *)
(*   i.e. before any use of skeleton expressions in the pardo application.   *)
(*****************************************************************************)

(*************************************************************)
(* Execution is controlled by data read in a config file     *)
(* This file should contain:                                 *)
(* - the name of the file defining the regular decomposition *)
(* - the basename for result files                           *)
(* - the output type: window, ps, none                       *)
(* - the number of iterations for the fix point problem      *)
(* - the number of iterations for the local subproblems      *)
(* - the refreshment step for the graphical device           *)
(* - the number of seconds to sleep before terminating       *)
(*************************************************************)

open Config;;
open Rdec2;;
open Gnuplot;;
open My_data;;

(* Management of the user options *)

let get_worker_dir, set_worker_dir =
  Uaref.make_ref_single "worker_dir" (Filename.concat ".." "worker")

and get_worker_filename, set_worker_filename =
  Uaref.make_ref_single "worker_filename" "poisson.deb"

and get_number_of_workers, set_number_of_workers =
  Uaref.make_ref_single "number_of_workers" 0

and get_config_dir, set_config_dir =
  Uaref.make_ref_single "config_dir" (Filename.concat ".." "config_files")

and get_config_filename, set_config_filename =
  Uaref.make_ref_single "config_filename" "config_default"

and get_res_dir, set_res_dir =
  Uaref.make_ref_single "res_dir" "res";;

let get_global_parameters () =
  let cfg_fname =
    Filename.concat (get_config_dir ()) (get_config_filename ()) in
  let cfg = parse_configuration_file cfg_fname in
  let rdec_fname =
    Filename.concat (get_config_dir ()) cfg.rdec_config_filename in
  let rdec = parse_rdec_config_file rdec_fname in
  check_regular_decomposition rdec;
  check_graphical_output cfg.graphical_output;
  if get_number_of_workers () = 0 then begin
    let nsbdx, nsbdy = number_of_subdomains_of_regular_decomposition rdec in
    set_number_of_workers (nsbdx * nsbdy) end;
  cfg_fname, cfg, rdec;;

(* The elements of the main data stream are of type subdomain array
   (see My_data). *)
let set_data () =
  (fun (cfg_fname, cfg, rdec) ->
    let intf0 = make_zero_interface rdec in
    let f i bd =
      {config_filename = cfg_fname;
       config = cfg;
       rdec = rdec;
       num = i;
       bd = bd;
       iter = 0;
       err = max_float, max_float} in
    Array.mapi f intf0);;

let init_poisson (_, num as sd) jtermax (_, oc) =
  let g = computational_rectangular_grid_of_subdomain sd in
  print_computational_rectangular_grid oc g;
  Printf.fprintf oc "%d\n" jtermax;
  flush oc;;

let computation_on_a_subdomain () =
  let bin = ref None and cout = ref None in
  (fun ({config = cfg; rdec = rd; num = n; bd = bd; iter = i} as subd) ->
    let i = i + 1 in
    let res = Filename.concat (get_res_dir ()) cfg.result_basename in
    let inmax = cfg.inner_itermax in
    let w_cmd = Filename.concat (get_worker_dir ()) (get_worker_filename ()) in
    let ib, oc =
      Uunix.spawn_init
        (init_poisson (rd, n) inmax) w_cmd bin cout in
    let go =
      match cfg.graphical_output with
      | No -> "."
      | _ ->
          if i mod cfg.display_step <> 0 then "." else
          Printf.sprintf "%s%i" res n in
    Printf.fprintf oc "%s\n" go;
    print_boundary oc bd;
    flush oc;
    let rel_stag, rel_err = Scanf.bscanf ib " %f %f\n" (fun e1 e2 -> e1, e2) in
    let bd = read_boundary ib (size_of_boundary bd) in
    {subd with bd = bd; iter = i; err = rel_stag, rel_err});;

let proj () =
  (fun v ->
    let rdec = get_rdec v in
    let table = connectivity_table_of_regular_decomposition rdec in
    let bds = get_bds v in
    let bds = projection table bds in
    set_bds bds v;
    v);;

let plot () =
  let bin = ref None and cout = ref None in
  (fun v ->
    let cfg = get_config v in
    let go = cfg.graphical_output in
    let bname = Filename.concat (get_res_dir ()) cfg.result_basename in
    begin match go with
    | No -> ()
    | _ ->
      let ic, oc = Uunix.spawn "gnuplot" bin cout in
      let iter = get_iter v in
      if iter mod cfg.display_step = 0 then begin
        let rdec = get_rdec v in
        let nsbd = number_of_subdomains_of_regular_decomposition rdec in
        let gpl = make_plot_command go bname iter nsbd in
        Printf.fprintf oc "%s\n" gpl;
        flush oc end; end;
    v);;

let convergence =
  let t0 = ref (Unix.gettimeofday ()) in
  (fun v ->
    let iter = get_iter v in
    let f (rs1, re1) {err = rs2, re2} = max rs1 rs2, max re1 re2 in
    let rel_stag, rel_err = Array.fold_left f (0. , 0.) v in
    let dt = Unix.gettimeofday () -. !t0 in
    Printf.eprintf
      "iter = %2d -> rel_stag = %5.0e, rel_err = %6.2f%% (time = %.2f s)\n"
      iter rel_stag (100. *. rel_err) dt;
    flush stderr;
    let cfg = get_config v in
    iter < cfg.outer_itermax);;

let print_result v =
  let fname = My_data.get_config_filename v in
  Printf.eprintf
    "\nComputation from configuration file %S completed.\n\n" fname;
  let cfg = get_config v in
  match cfg.graphical_output with
  | Window -> Unix.sleep cfg.a_while
  | _ -> ();;
