module Client (Streams: Streams.Type) : P3l.Implem = struct

  module Opt = Opt.Init(struct end)

  module Streams = Streams

  type color = int
      
  type 'a stream = 'a Streams.t

  open Streams

  let port = Opt.port
    
  let new_id =
    let cpt = ref 0 in
    fun () -> incr cpt; "seq_"^(string_of_int !cpt)
				  
  let register id f =
    def f_chan(x) = reply f x to f_chan in
    Join.Ns.register Join.Ns.here id f_chan 
      
  type ('a, 'b) skel = unit

  let seq ?col:color f = 
    let id = new_id () in
    let _ = register id f in
Join_utils.print_DEBUG ("Distributed.Client.seq : register "^id);
    ()
      
  let farm ?col:color ?colv:color_list (skl, n) = ()
    
  let mapvector ?col:color ?colv:color_list (skl, n) = ()
      
  let pipeline skl1 skl2 = ()
      
  let ( ||| ) skl1 skl2 = pipeline skl1 skl2
      
  let parfun f = fun _ -> assert false


  let loop ?col:c (p, skl) = (() : ('a, 'b) skel)
    
  let pardo f = 
    Join.Site.listen (Unix.ADDR_INET (Join.Site.get_local_addr(), port));
    def wait() & abs() = reply to wait in
    wait()

end

module Server (Streams: Streams.Type) : P3l.Implem = struct

  module Opt_root = Opt_root.Init(struct end)

  module Streams = Streams

  type color = int
      
  type 'a stream = 'a Streams.t

  open Streams

  type ('a, 'b) skel = 'a Streams.t -> 'b Streams.t
(* A skeleton is a function from lazy stream of A to lazy stream of B. *)

  let do_plug f c s = f (c (fork s))
(* [do_plug f c s] plugs a parallel circuit with 2 inputs and two outputs to
    two wires (whith a [fork] emitter and a [f] collector at both ends). *)

  let plug c = do_plug join c
(* [plug c] returns a circuit that fork a stream [s] and applies [c] to those
    two streams, then join the results into a resulting stream.
    In short, [plug] plugs a parallel circuit with 2 inputs and two outputs to
    two wires (whith an emitter and a collector at both ends).
    [plug c s] = [join (c (fork s))]. *)

  let par_apply skl1 skl2 (s1, s2) = 
    def r1(x1) & r2(x2) & wait() = reply (x1, x2) to wait 
    in
    spawn r1(skl1 s1);
    spawn r2(skl1 s2);
    wait()
(* Applies two skeletons in parallel onto 2 streams. *)

  let in_par skl = par_apply skl skl
(* Applies one skeleton onto two streams in parallel. *)

  let seq_apply to_stream of_stream skl x = of_stream (skl (to_stream x))
(* Turns a skeleton from lazy stream of A to lazy stream of B into
   a regular function from A to B, by creating a stream from the
   regular argument and extracting a regular value from the resulting
   stream. *)

  let in_farm collect (skl, n) =
    if n <= 0 then failwith "farm: non positive number of workers." else
    let rec in_farm (skl, n) = 
      match n with
      | 1 -> skl
      | n ->
	  let r = n mod 2
	  and q = n / 2 in
	  match r with
	  | 0 ->
              collect (in_par (in_farm (skl, q)))
	  | _ ->
              collect
		(par_apply
		   (in_farm (skl, q + 1))
		   (in_farm (skl, q))) in
    in_farm (skl, n)
(* Launches n copies of [skl] and collects the results using function
   [collect]. *)

  let new_id =
    let cpt = ref 0 in
    fun () -> incr cpt; "seq_"^(string_of_int !cpt)
				  
  let create_pool () =
    def compute(x) & agent(f) =
      let r = f x in
      agent(f) & reply r to compute
    in
    agent, compute

  let seq ?col:color f = 
    let id = new_id () in
    let reg, compute = create_pool () in
Join_utils.print_DEBUG ("Distributed.Server.seq : register "^id);
    let _ = 
      List.iter
	(fun ns -> 
	  let (f: 'a -> 'b) = Join_utils.lookup ns id in
	  spawn (reg f))
	Opt_root.clients_ns
    in
    Streams.map compute

  let unseq skl = seq_apply singleton head skl

  let farm ?col:color ?colv:color_list (skl, n) = in_farm plug (skl, n)

  let mapvector ?col:color ?colv:color_list (skl, n) =
    let fskl = farm (skl, n) in
    let f = seq_apply Streams.of_vect Streams.to_vect fskl in
    Streams.map f

  type 'a buffer = 
      { put : 'a -> unit ; 
	get : unit -> 'a ;
	end_of_stream : unit Join.chan; }


  let create_buffer (size) =
    def state(xs, ys, Some n) & put(x) = 
      begin 
	if n > 1 then
	  state(x::xs, ys, Some (n-1)) 
	else
	  state(x::xs, ys, None)
      end
	& reply to put 
    or state (xs, y::ys, Some n) & get() = 
      state(xs, ys, Some (n+1)) & reply y to get
    or state (xs, y::ys, None) & get() = 
      state(xs, ys, Some 1) & reply y to get
    or state(_::_ as xs, [], free_cpt) & get() = 
      state([], List.rev xs, free_cpt) & reply get() to get
    or state([], [], free_cpt) & put(x) & get() =
      state([], [], free_cpt) & reply to put & reply x to get
    or state([], [], _) & end_of_stream() & get() =
      reply raise End_of_stream to get
    in
    spawn begin match size with 
    | 0 -> state([],[],None) 
    | n when n > 0 -> state([],[],Some n)
    | _ -> state([],[],Some max_int)
    end;
    {put=put; get=get; end_of_stream = end_of_stream; }
 

  let bufferize s n =
    let buff = create_buffer (n) in
    def transmit() = 
      let x_opt =
	try Some (head s)
	with End_of_stream -> None
      in
      match x_opt with
      | Some x -> buff.put(x) ; transmit() 
      | None -> buff.end_of_stream()
    in
    spawn transmit() ;
    of_fun buff.get


  let pipeline skl1 skl2 s =
    let s' = skl1 s in
    let bufferized_s' = bufferize s' 1 in
    skl2 bufferized_s'


  let ( ||| ) skl1 skl2 = pipeline skl1 skl2


  let loop ?col:c (p, skl) s =
    let rec loop s =
      try 
	let (y, t) = Streams.head_tail (skl s) in
	if p y 
	then loop (Streams.scons y t)
	else Streams.scons y (loop t)
      with
      | Streams.End_of_stream -> Streams.snil
    in
    (loop s)

  let parfun f = f
  
  let pardo f = 
    f ()

end
