(* $Id: seq.ml,v 1.2 2009/06/25 15:08:06 mandel Exp $ *)

module Implem : P3l.Implem = struct

module Streams = Streams_lazy

type color = int

type 'a stream = 'a Streams.t

type ('a, 'b) skel = 'a stream -> 'b stream;;
(* A skeleton is a function from lazy stream of A to lazy stream of B. *)

open Streams;;


let do_plug f c s = f (c (fork s));;
(* [do_plug f c s] plugs a parallel circuit with 2 inputs and two outputs to
    two wires (whith a [fork] emitter and a [f] collector at both ends). *)

let plug c = do_plug join c;;
(* [plug c] returns a circuit that fork a stream [s] and applies [c] to those
    two streams, then join the results into a resulting stream.
    In short, [plug] plugs a parallel circuit with 2 inputs and two outputs to
    two wires (whith an emitter and a collector at both ends).
    [plug c s] = [join (c (fork s))]. *)

let par_apply skl1 skl2 (s1, s2) = (skl1 s1, skl2 s2);;
(* Applies two skeletons in parallel onto 2 streams. *)

let in_par skl = par_apply skl skl;;
(* Applies one skeleton onto two streams in parallel. *)

let seq_apply to_stream of_stream skl x = of_stream (skl (to_stream x));;
(* Turns a skeleton from lazy stream of A to lazy stream of B into
   a regular function from A to B, by creating a stream from the
   regular argument and extracting a regular value from the resulting
   stream. *)

let in_farm collect (skl, n) =
  if n <= 0 then failwith "farm: non positive number of workers." else
  let rec in_farm (skl, n) = 
    match n with
    | 1 -> skl 
    | n ->
      let r = n mod 2
      and q = n / 2 in
      match r with
      | 0 ->
        collect (in_par (in_farm (skl, q)))
      | _ ->
        collect
          (par_apply
             (in_farm (skl, q + 1))
             (in_farm (skl, q))) in
  in_farm (skl, n);;
(* Launches n copies of [skl] and collects the results using function
   [collect]. *)

let seq ?col:color f = Streams.map f;;

let unseq skl = seq_apply singleton head skl;;

let farm ?col:color ?colv:color_list (skl, n) = in_farm plug (skl, n);;

let mapvector ?col:color ?colv:color_list (skl, n) =
  let fskl = farm (skl, n) in
  let f = seq_apply Streams.of_vect Streams.to_vect fskl in
  Streams.map f;;

let stream_array_of_array_stream : ('a array stream -> 'a stream array) =
  fun s -> assert false

let array_stream_of_stream_array : ('a stream array -> 'a array stream) =
  fun s -> assert false


let par_apply (skl_array: ('a, 'b) skel array) array_stream =
  let stream_array = stream_array_of_array_stream array_stream in
  array_stream_of_stream_array
    (Array.mapi 
       (fun i x -> skl_array.(i) x)
       stream_array)

(*   Array.init *)
(*     (Array.length stream_array) *)
(*     (fun i -> skl_array.(i) stream_array.(i)) *)

(* 
let rec reduce_farm e (f, n) =
  let c = fold_left f e in
  let plug = do_plug (fun (x, y) -> f x y) in
  in_farm plug (c, n);;

let reducevector e (acf, n) =
  let skl = reduce_farm e (acf, n) in
  let f v = skl (Streams.of_vect v) in
  Streams.map f;;
*)

let loop ?col:c (p, skl) s =
  let rec loop s =
    match Lazy.force (skl s) with
    | Nil -> Nil
    | Cons (y, t) ->
      if p y
      then loop (lazy (Cons (y, t)))
      else Cons (y, lazy (loop t)) in
  lazy (loop s);;


let pipeline skl1 skl2 s = 
  lazy (Lazy.force (skl2 (skl1 s)));;

let ( ||| ) skl1 skl2 = pipeline skl1 skl2;;

let parfun f = f;;

let pardo f = f ();;

end
