(* $Id: streams.ml,v 1.7 2007/02/25 21:54:57 weis Exp $ *)

(* Streams:
 No thunk on the wires (only plain values are transmitted)
 => no lazy values in the streams.
 Lazy evaluation in networks (meaning next elements is on demand)
 => lazy evaluation in the streams. *)

open Lazy;;

exception End_of_stream;;

(* Lazy stream, which are lazy in tail. *)
type 'a strict_stream =
   | Nil
   | Cons of 'a * 'a stream

and 'a stream = 'a strict_stream Lazy.t;;

(* [snil] is ``stream nil'', the empty stream. *)
let snil = lazy (
  Nil
);;

(* [scons] is ``stream cons'' the lazy stream constructor;
   [scons] is hard to use, since it is impossible to prevent the evaluation of
   the [s] argument, if it is not simply a variable in the callee. *)
let scons x s = lazy (
  Cons (x, s)
);;

let head s =
  match force s with
  | Cons (x, _) -> x
  | Nil -> raise End_of_stream;;

let singleton x = scons x snil;;

(* [iter f s] iterates the function [f] on the stream [s]:
   if [s] is [x1; x2; ...; xn], then [iter f s] performs the sequence
   [f x1; f x2; ...; f xn].
   Note: any call to [iter] completely unfreeezes its the stream argument. *)
let rec iter f s =
  match force s with
  | Nil -> ()
  | Cons (x, t) -> f x; iter f t;;

(* [fold_left f e s] folds the binary function [f] on the stream [s]:
   if [s] is [x1; x2; ...; xn], then [fold_left f e s] computes
   [f (... (f (f e x1) x2) ...) xn].
   Note: any call to [fold_left] completely unfreeezes its the stream argument. *)
let rec fold_left f accu s =
  match force s with
  | Nil -> accu
  | Cons (x, t) -> fold_left f (f accu x) t;;

(* [map f s] maps the function [f] on the stream [s]:
   if [s] is [x1; x2; ...; xn], then [map f s] returns the stream
   [f x1; f x2; ...; f xn]. *)
let rec map f s = lazy (
  match force s with
  | Nil -> Nil
  | Cons (x, t) -> Cons (f x, map f t));;

(* Generate a lazy stream from a function. *)
let rec of_fun f = lazy (
  match
   try Some (f ()) with
   | End_of_stream -> None
  with
  | None -> Nil
  | Some x -> Cons (x, of_fun f));;

(* Generate a lazy stream from a list. *)
let rec of_list l = lazy (
  match l with
  | [] -> Nil
  | x :: t -> Cons (x, of_list t)
);;

(* Generate the list of elements of a lazy stream.
   Warning: this function is eager, meaning that the stream argument is
   completely evaluated (if this is possible, otherwise the function loops for
   ever). *)
let to_list s =
  let rec loop accu = function
  | Nil -> accu
  | Cons (x, t) -> loop (x :: accu) (force t) in
  List.rev (loop [] (force s));;

(* Generate a stream from a vector. *)
let of_vect v = lazy (
  match v with
  | [||] -> Nil
  | v ->
    let len = Array.length v in
    let rec loop i =
      if i < len then Nil else
      Cons (v.(i), lazy (loop (i + 1))) in
    loop 0
);;

(* Generate the vector of elements of a lazy stream.
   Warning: this function is eager, meaning that the stream argument is
   completely evaluated (if this is possible, otherwise the function loops for
   ever). *)
let to_vect s = Array.of_list (to_list s);;

(* Merges two lazy stream into a single stream with intermixed elements from
   both streams in a fair way (alternatively from one at a time). *)
let rec join (s1, s2) = lazy (
  match force s1 with
  | Nil -> force s2
  | Cons (x, t) -> Cons (x, join (s2, t))
);;

(* [join_reduce f (s1, s2)]:
   [f] is a binary function,
   [s1] and [s2] are two streams with elements of type ['a].
   [join_reduce] is a kind of [map2] functional.
   If [s1] is [x1; x2; ...; xn] and [s2] is [y1; y2; ...; yn], then
   [join_reduce f (s1. s2)] is [f x1 y1; f x2 y2; ...; f xn yn].
   Operations with [f] stops when one of the stream is exhausted;
   in this case, the rest of the non exhausted stream is the rest of the
   result. *)
(*let rec join_reduce f (s1, s2) = lazy (
  match force s1 with
  | Nil -> force s2
  | Cons (x1, t1) as fs1 ->
    match force s2 with
    | Nil -> fs1
    | Cons (x2, t2) ->
      Cons (f x1 x2, join_reduce f (t1, t2))
);;*)

let rec join_pair (s1, s2) = lazy (
  match force s1 with
  | Nil -> begin
    match force s2 with
    | Nil -> Nil
    | _ -> failwith "join_pair: first stream too short." end
  | Cons (x1, t1) -> begin
    match force s2 with
    | Nil -> failwith "join_pair: second stream too short."
    | Cons (x2, t2) ->
      Cons ((x1, x2), join_pair (t1, t2)) end
);;

(* [fork s] separates the stream [s] into two streams in a fair way, even
   order elements are in one stream, odd order are in the other.
   (This means in particular, that if [s] is finite and has length [l] the two
   streams have length [l/2] is [l] is even, and respectively length [l/2] and
   [l/2 + 1] if [l] is odd .) *)
let rec fork s =
  let fs = force s in
  match fs with
  | Nil -> snil, snil
  | Cons (x, t) ->
    match force t with
    | Nil -> s, snil
    | Cons (y, t) ->
      (* Sharing the rest of the resulting (lazy) pair of streams. *)
      let pair = lazy (fork t) in
      (lazy (Cons (x, (lazy (force (fst (force pair)))))),
       lazy (Cons (y, (lazy (force (snd (force pair)))))))
;;

let rec fair_fork a s =
  let fs = force s in
  match fs with
  | Nil -> snil, snil
  | Cons (x, t) ->
    match force t with
    | Nil -> s, singleton a
    | Cons (y, t) ->
      (* Sharing the rest of the resulting (lazy) pair of streams. *)
      let pair = lazy (fair_fork a t) in
      (lazy (Cons (x, (lazy (force (fst (force pair)))))),
       lazy (Cons (y, (lazy (force (snd (force pair)))))))
;;

type 'a t = 'a stream;;
