(* $Id: streams.mli,v 1.6 2007/02/25 21:54:57 weis Exp $ *)

(** Streams are used to modelize the behaviour of computational nodes.

 Computational nodes are a kind of light servers or processes or threads, that get
 there input from the wires on the network and send their results on the wires.

 Since there is no thunk on the wires, there are no lazy values in the streams.
 Networks feature delay in values availability, hence there is a kind of lazy
 evaluation in networks that we modelize via lazy evaluation in the streams.

 In networks, even to get the first element of a stream, we have to explicitely
 ask and wait; hence a stream is a lazy data structure that has to be forced to get
 its head or even to know that it is indeed empty.
*)

(** Lazy stream, which are lazy in tail. *)
type 'a strict_stream =
   | Nil
   | Cons of 'a * 'a stream

and 'a stream = 'a strict_stream Lazy.t;;

(** Construction functions. *)
val snil : 'a stream;;
val scons : 'a -> 'a stream -> 'a stream;;
val singleton : 'a -> 'a stream;;
(** [singleton x] returns the stream that has only [x] as element. *)

val head : 'a stream -> 'a;;
(** [head s] returns the first element of stream [s].
  Raises [End_of_stream], if [s] is empty. *)

(** Generate a lazy stream from a function. *)
val of_fun : (unit -> 'a) -> 'a stream;;

exception End_of_stream;;

(** Generate a lazy stream from a list. *)
val of_list : 'a list -> 'a stream;;

(** Generate a list from a lazy stream. *)
val to_list : 'a stream -> 'a list;;

(** Generate a lazy stream from a vector. *)
val of_vect : 'a array -> 'a stream;;

(** Generate the vector of the elements of a (finite) lazy stream. *)
val to_vect : 'a stream -> 'a array;;

(** Operations on streams. *)

(** Merges two lazy stream into a single stream with intermixed elements from
    both streams in a fair way (fairly peeks up alternatively elements from
    both streams). *)
val join : ('a stream * 'a stream) -> 'a stream;;

(** Splits a stream into two. *)
val fork : 'a stream -> ('a stream * 'a stream);;

val iter : ('a -> unit) -> 'a stream -> unit;;
val map : ('a -> 'b) -> 'a stream -> 'b stream;;
val fold_left : ('a -> 'b -> 'a) -> 'a -> 'b stream -> 'a;;

type 'a t = 'a stream;;
